from __future__ import annotations

from escnn.group import change_basis, directsum
from escnn.group.irrep import generate_irrep_matrices_from_generators
from escnn.group.irrep import restrict_irrep
from escnn.group.utils import cycle_isclose

from .utils import *

from .so3_utils import PARAMETRIZATION as PARAMETRIZATION_SO3
from .so3_utils import PARAMETRIZATIONS
from .so3_utils import IDENTITY, _grid, _combine, _equal, _invert, _change_param, _check_param, _hash

from .so3group import _build_character, _build_irrep

import numpy as np

from typing import Tuple, Callable, Iterable, List, Dict, Any, Union


__all__ = ["Icosahedral"]

_PHI = (1. + np.sqrt(5)) / 2


class Icosahedral(Group):

    PARAM = PARAMETRIZATION_SO3
    
    PARAMETRIZATIONS = PARAMETRIZATIONS
    
    def __init__(self):
        r"""
        
        Subgroup Structure:

        +-----------------------------------+-----------------------------------+-------------------------------------------------------------------------------------------------------------------+
        |    ``id[0]``                      |    ``id[1]``                      |    subgroup                                                                                                       |
        +===================================+===================================+===================================================================================================================+
        |        'ico'                      |                                   |   The Icosahedral :math:`I` group itself                                                                          |
        +-----------------------------------+-----------------------------------+-------------------------------------------------------------------------------------------------------------------+
        |        'tetra'                    |                                   |   Tetrahedral :math:`T` subgroup                                                                                  |
        +-----------------------------------+-----------------------------------+-------------------------------------------------------------------------------------------------------------------+
        |        False                      |     N = 1, 2, 3, 5                |   :math:`C_N` of N discrete planar rotations                                                                      |
        +-----------------------------------+-----------------------------------+-------------------------------------------------------------------------------------------------------------------+
        |        True                       |     N = 2, 3, 5                   |   *dihedral* :math:`D_N` subgroup of N discrete planar rotations and out-of-plane :math:`\pi` rotation            |
        +-----------------------------------+-----------------------------------+-------------------------------------------------------------------------------------------------------------------+
        |        True                       |     1                             |   equivalent to ``(False, 2, adj)``                                                                               |
        +-----------------------------------+-----------------------------------+-------------------------------------------------------------------------------------------------------------------+

        """
        
        super(Icosahedral, self).__init__("Icosahedral", False, False)
        
        self._identity = self.element(IDENTITY)
        
        self._elements = [self.element(g) for g in _grid('ico')]
        assert len(self._elements) == 60
            
        # self._identity = self._elements[3]

        self._generators = [
            self._elements[21],                           # Cyclic Group of order 5
            self._elements[0],                            # Cyclic Group of order 2
            # self._elements[38]                          # Cyclic Group of order 3
            # self._elements[0] @ self._elements[21]      # Cyclic Group of order 3
        ]
        
        self._build_representations()

    @property
    def generators(self) -> List[GroupElement]:
        return self._generators

    @property
    def identity(self) -> GroupElement:
        return self._identity

    @property
    def elements(self) -> List[GroupElement]:
        return self._elements
     
    @property
    def _keys(self) -> Dict[str, Any]:
        return dict()

    @property
    def subgroup_trivial_id(self):
        return (False, 1)

    @property
    def subgroup_self_id(self):
        return 'ico'

    ###########################################################################
    # METHODS DEFINING THE GROUP LAW AND THE OPERATIONS ON THE GROUP'S ELEMENTS
    ###########################################################################

    def _inverse(self, element, param=PARAM):
        r"""
        Return the inverse element of the input element
        """
        return _invert(element, param=param)

    def _combine(self, e1, e2,
                param=PARAM,
                param1=None,
                param2=None
                ):
        r"""
        Return the sum of the two input elements
        """
        return _combine(e1, e2, param=param, param1=param1, param2=param2)

    def _equal(self, e1, e2,
              param=PARAM,
              param1=None,
              param2=None,
              ) -> bool:
        r"""
        Check if the two input values corresponds to the same element
        """
        return _equal(e1, e2, param=param, param1=param1, param2=param2)

    def _hash_element(self, element, param=PARAM):
        return _hash(element, param)

    def _repr_element(self, element, param=PARAM):
        return element.__repr__()

    def _is_element(self, element,
                    param: str = PARAM,
                    verbose: bool = False,
                    ) -> bool:
    
        ATOL = 1e-7
        RTOL = 1e-5
        angle_ATOL = 1e-6
        
        if not _check_param(element, param):
            if verbose:
                print(f"Element {element} is not a rotation")
            return False
    
        element = self._change_param(element, param, 'Q')
        
        v = element[:3]
        theta = 2*np.arccos(np.clip(element[3], -1., 1.))
        
        if cycle_isclose(theta, 0., 2*np.pi, atol=angle_ATOL, rtol=0.):
            return True
        
        v = v / np.sin(theta/2.)
        
        if cycle_isclose(theta, 0., 2 * np.pi / 2, atol=angle_ATOL, rtol=0.):
            # rotation of order 2
            return _is_axis_aligned(v, 2, verbose=verbose, ATOL=ATOL, RTOL=RTOL)

        elif cycle_isclose(theta, 0., 2 * np.pi / 3, atol=angle_ATOL, rtol=0.):
            # rotation or order 3
            return _is_axis_aligned(v, 3, verbose=verbose, ATOL=ATOL, RTOL=RTOL)

        elif cycle_isclose(theta, 0., 2 * np.pi / 5, atol=angle_ATOL, rtol=0.):
            # rotation or order 5
            return _is_axis_aligned(v, 5, verbose=verbose, ATOL=ATOL, RTOL=RTOL)

        else:
            if verbose:
                print(f'Group element is neither a 2-fold, a 3-fold nor a 5-fold rotation.')
            return False

    def _change_param(self, element, p_from: str, p_to: str):
        assert p_from in self.PARAMETRIZATIONS
        assert p_to in self.PARAMETRIZATIONS
        return _change_param(element, p_from, p_to)

    ###########################################################################

    def sample(self, param: str = PARAM) -> GroupElement:
        return self._elements[
            np.random.randint(self.order())
        ]

    def testing_elements(self) -> Iterable[GroupElement]:
    
        r"""
        A finite number of group elements to use for testing.
        """
        return iter(self._elements)

    def change_param(self, element, p_from, p_to):
        return _change_param(element, p_from, p_to)

    def __eq__(self, other):
        if not isinstance(other, Icosahedral):
            return False
        else:
            return self.name == other.name

    def _process_subgroup_id(self, id):

        if not isinstance(id, tuple):
            id = (id,)

        assert isinstance(id[0], bool) or isinstance(id[0], str), id[0]

        if not isinstance(id[-1], GroupElement):
            id = (*id, self.identity)

        assert id[-1].group == self

        if isinstance(id[0], bool):
            assert id[1] in [1, 2, 3, 5]

            if id[0] == True and id[1] == 1:
                # flip subgroup of the O(2) subgroup of SO(3)
                # this is equivalent to the C_2 subgroup of 180 deg rotations out of the plane (around X axis)

                # V = np.asarray([0., -_PHI, 1 / _PHI])
                V = np.array([1., 1., -1.])
                V /= np.linalg.norm(V)

                change_axis = np.zeros(4)
                change_axis[:3] = V * np.sin(np.pi/3.)
                change_axis[3] = np.cos(np.pi/3.)

                adj = self.element(change_axis, 'Q') @ id[-1]
                id = (False, 2, adj)

        return id

    def _subgroup(self, id) -> Tuple[
        Group,
        Callable[[GroupElement], GroupElement],
        Callable[[GroupElement], GroupElement]
    ]:
        r"""
        
        Returns:
            a tuple containing

                - the subgroup,

                - a function which maps an element of the subgroup to its inclusion in the original group and

                - a function which maps an element of the original group to the corresponding element in the subgroup (returns None if the element is not contained in the subgroup)
              
        """
        # TODO : implement this!

        sg = None
        parent_map = None
        child_map = None

        id, adj = id[:-1], id[-1]

        if id == ('ico',):
            sg = self
            parent_map = build_adjoint_map(self, ~adj)
            child_map = build_adjoint_map(self, adj)
        elif id == ('tetra',):
            raise NotImplementedError()
        elif id == (False, 1):
            sg = escnn.group.cyclic_group(1)
            parent_map, child_map = build_trivial_subgroup_maps(self)
        elif id == (False, 2):
            sg = escnn.group.cyclic_group(2)
            axis = np.asarray([0., 0., 1.])
            parent_map = cn_to_ico(adj, sg, axis=axis)
            child_map = ico_to_cn(adj, sg, axis=axis)
        elif id == (False, 3):
            sg = escnn.group.cyclic_group(3)
            axis = np.asarray([1., 1., 1.]) / np.sqrt(3)
            parent_map = cn_to_ico(adj, sg, axis=axis)
            child_map = ico_to_cn(adj, sg, axis=axis)
        elif id == (False, 5):
            sg = escnn.group.cyclic_group(5)
            axis = np.asarray([_PHI, 0., 1.])
            axis /= np.linalg.norm(axis)
            parent_map = cn_to_ico(adj, sg, axis=axis)
            child_map = ico_to_cn(adj, sg, axis=axis)
        elif id == (True, 2):
            sg = escnn.group.dihedral_group(2)
            parent_map, child_map = None, None
            raise NotImplementedError()
        elif id == (True, 3):
            sg = escnn.group.dihedral_group(3)
            parent_map, child_map = None, None
            raise NotImplementedError()
        elif id == (True, 5):
            sg = escnn.group.dihedral_group(5)
            parent_map, child_map = None, None
            raise NotImplementedError()
        else:
            raise ValueError(f'Subgroup id {id} not recognized!')

        return sg, parent_map, child_map

    def _restrict_irrep(self, irrep: str, id) -> Tuple[np.matrix, List[str]]:
        r"""
        
        Returns:
            a pair containing the change of basis and the list of irreps of the subgroup which appear in the restricted irrep
            
        """
        sg_id, adj = id[:-1], id[-1]

        irr = self.irrep(*irrep)

        sg, _, _ = self.subgroup(id)

        irreps = []
        change_of_basis = None

        try:
            if sg_id == ('ico', ):
                change_of_basis = irr.change_of_basis
                irreps = irr.irreps
            elif sg_id == (False, 1):
                change_of_basis = np.eye(irr.size)
                irreps = [(0,)]*irr.size
            else:
                raise NotImplementedError()

        except NotImplementedError:
            if sg.order() > 0:
                change_of_basis, irreps = restrict_irrep(irr, sg_id)
            else:
                raise

        change_of_basis = self.irrep(*irrep)(adj).T @ change_of_basis

        return change_of_basis, irreps

    def _build_representations(self):
        r"""
        Build the irreps for this group

        """

        # Build all the Irreducible Representations

        # add Trivial representation
        self.irrep(0)

        # add other irreducible representations
        self.irrep(1)
        self.irrep(2)

        # SO(3)'s freq 3 irrep decomposes in another 3 dimensional irrep and a 4 dimensional one
        self.irrep(3)
        self.irrep(4)

        # add all the irreps to the set of representations already built for this group
        self.representations.update(**{irr.name: irr for irr in self.irreps()})

        # build the regular representation

        # N.B.: it represents the LEFT-ACTION of the elements
        self.representations['regular'] = self.regular_representation

    @property
    def trivial_representation(self) -> Representation:
        return self.irrep(0)

    @property
    def standard_representation(self) -> Representation:
        r"""
        Restriction of the standard representation of SO(3) as 3x3 rotation matrices

        """
        name = f'standard'
    
        if name not in self._representations:
            change_of_basis = np.array([
                [0, 0, 1],
                [1, 0, 0],
                [0, 1, 0]
            ])
        
            self._representations[name] = change_basis(
                self.irrep(1),
                change_of_basis=change_of_basis,
                name=name,
                supported_nonlinearities=self.irrep(1).supported_nonlinearities,
            )
    
        return self._representations[name]

    @property
    def ico_vertices_representation(self) -> Representation:
        # action on the 12 vertices of the icosahedron (or faces of the dodecahedron)
        
        # quotient repr wrt C_5 subgroup?
        return self.quotient_representation((False, 5), name='ico_vertices')

    @property
    def ico_faces_representation(self) -> Representation:
        # action on the 20 faces of the icosahedron (or vertices of the dodecahedron)
        
        # quotient repr wrt C_3 subgroup?
        return self.quotient_representation((False, 3), name='ico_faces')

    @property
    def ico_edges_representation(self) -> Representation:
        # action on the 30 edges of the icosahedron or dodecahedron
        
        # quotient repr wrt C_2 subgroup

        # n.b.: C_2 is the symmetry group of an edge

        return self.quotient_representation((False, 2), name='ico_edges')

    def bl_regular_representation(self, L: int) -> Representation:
        r"""
        Band-Limited regular representation up to frequency ``L`` (included).

        Args:
            L(int): max frequency

        """

        assert isinstance(L, int)
        assert 0 <= L <= 4

        name = f'regular_{L}'

        if name not in self._representations:
            irreps = []

            for l in range(L + 1):
                irr = self.irrep(l)
                irreps += [irr] * irr.size

            self._representations[name] = directsum(irreps, name=name)

        return self._representations[name]

    def irrep(self, l: int) -> IrreducibleRepresentation:
        r"""
        Build the irrep of :math:`I` identified by the non-negative integer :math:`l`.
        For :math:`l = 0, 1, 2`, the irrep is equivalent to the Wigner D matrix of the same frequency :math:`l`.
        For :math:`l=3`, the 7-dimensional Wigner D matrix is decomposed in a 3-dimensional and a 4-dimensional irrep,
        here identified respectively by :math:`l=3` and :math:`l=4`.
        
        Args:
            l (int): identifier of the irrep

        Returns:
            the corresponding irrep

        """
        
        assert isinstance(l, int)
        assert 0 <= l <= 4
        name = f"irrep_{l}"
        id = (l,)

        if id not in self._irreps:

            if l == 0:
                # Trivial representation
                irrep = build_trivial_irrep()
                character = build_trivial_character()
                supported_nonlinearities = ['pointwise', 'norm', 'gated', 'gate']
                self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, 1, 'R',
                                                              supported_nonlinearities=supported_nonlinearities,
                                                              character=character,
                                                              frequency=0
                                                              )
            elif l <= 2:
        
                # other Irreducible Representations which are equivalent to Wigner D matrices
                # irrep = lambda element, l=l: _wigner_d_matrix(element.to(element.param), l=l, param=element.param)
                # character = lambda element, l=l: _character(element.to(element.param), l=l, param=element.param)
                irrep = _build_irrep(l)
                character = _build_character(l)
                supported_nonlinearities = ['norm', 'gated']
                self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, 2*l+1, 'R',
                                                              supported_nonlinearities=supported_nonlinearities,
                                                              character=character,
                                                              frequency=l)
            elif l == 3 or l == 4:
    
                irrep = _build_ico_irrep(self, l)
                supported_nonlinearities = ['norm', 'gated']
                self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, irrep[self.identity].shape[0], 'R',
                                                             supported_nonlinearities=supported_nonlinearities,
                                                             frequency=l)

            else:
                raise ValueError()

        return self._irreps[id]

    _cached_group_instance = None

    @classmethod
    def _generator(cls) -> 'Icosahedral':
        if cls._cached_group_instance is None:
            cls._cached_group_instance = Icosahedral()
    
        return cls._cached_group_instance


def _is_axis_aligned(v: np.ndarray, n: int, verbose: bool = False, ATOL=1e-7, RTOL = 1e-5) -> bool:

    norm = np.linalg.norm(v)
    v = v / norm

    if n == 2:
        # rotation of order 2

        # the rotation axis need to pass through the center of an edge of the icosahedron or the dodecahedron
        # These 30 points can be found as any cyclic permutation or change of sign of each element of
        # these 2 vectors:
        p = np.array([1., 0., 0.])  # 6 combinations
        q = np.array([_PHI + 1, _PHI, 1.])  # 24 combinations
        q /= np.linalg.norm(q)

        # remove sign ambiguity
        v = np.abs(v)
        # fix a permutation making the highest value first
        m = np.argmax(v)
        v = np.roll(v, -m)

        ans = np.allclose(v, p, atol=ATOL, rtol=RTOL) or np.allclose(v, q, atol=ATOL, rtol=RTOL)

        if not ans and verbose:
            print(f'Rotation by a multiple of pi/2 not aligned with a 2-fold rotational axis of the Icosahedron.')

        return ans

    elif n == 3:
        # rotation or order 3

        # the rotation axis need to pass through a vertex of the dodecahedron
        # These 20 points can be found as any cyclic permutation or change of sign of each element of
        # these 2 vectors:

        p = np.array([1., 1., 1.])  # 8 combinations
        q = np.array([_PHI, 1 / _PHI, 0.])  # 12 combinations

        p /= np.linalg.norm(p)
        q /= np.linalg.norm(q)

        # remove sign ambiguity
        v = np.abs(v)
        # fix a permutation making the highest value first
        m = np.argmax(v)
        v = np.roll(v, -m)

        ans = np.allclose(v, p, atol=ATOL, rtol=RTOL) or np.allclose(v, q, atol=ATOL, rtol=RTOL)

        if not ans and verbose:
            print(f'Rotation by a multiple of 2pi/3 not aligned with a 3-fold rotational axis of the Icosahedron.')

        return ans

    elif n == 5:
        # rotation or order 5

        # the rotation axis need to pass through a vertex of the icosahedron
        # These 12 points can be found as any cyclic permutation or change of sign of each element of this vector:
        p = np.array([_PHI, 0., 1.])  # 12 combinations
        p /= np.linalg.norm(p)

        # remove sign ambiguity
        v = np.abs(v)
        # fix a permutation making the highest value first
        m = np.argmax(v)
        v = np.roll(v, -m)

        ans = np.allclose(v, p, atol=ATOL, rtol=RTOL)

        if not ans and verbose:
            print(f'Rotation by a multiple of 2pi/5 not aligned with a 5-fold rotational axis of the Icosahedron.')

        return ans

    else:
        raise  ValueError('The rotation order must be one of {2, 3, 5}.')


#############################################
# SUBGROUPS MAPS
#############################################

# C_N #####################################

def ico_to_cn(adj: GroupElement, cn: escnn.group.CyclicGroup, axis: np.ndarray):
    assert isinstance(adj.group, Icosahedral)

    assert axis.shape == (3,)
    assert np.isclose(np.linalg.norm(axis), 1.)

    assert cn.order() in [2, 3, 5]

    assert _is_axis_aligned(axis, cn.order())

    def _map(e: GroupElement, cn=cn, adj=adj, axis=axis):
        ico = adj.group
        assert e.group == ico

        e = adj @ e @ (~adj)

        e = e.to('Q')

        v = e[:3]

        n = np.linalg.norm(v)

        if np.allclose(n, 0.):
            return cn.identity
        elif np.allclose(v / n, axis):
            # if the rotation is along the axis
            s, c = n, e[-1]
            theta = 2 * np.arctan2(s, c)
            try:
                return cn.element(theta, 'radians')
            except ValueError:
                return None
        else:
            return None

    return _map


def cn_to_ico(adj: GroupElement, cn: escnn.group.CyclicGroup, axis: np.ndarray):
    assert isinstance(adj.group, Icosahedral)

    assert axis.shape == (3,)
    assert np.isclose(np.linalg.norm(axis), 1.)

    assert cn.order() in [2, 3, 5]

    assert _is_axis_aligned(axis, cn.order())

    def _map(e: GroupElement, cn=cn, adj=adj, axis=axis):
        assert e.group == cn
        ico = adj.group

        theta_2 = e.to('radians') / 2.

        q = np.empty(4)
        q[:3] = axis * np.sin(theta_2)
        q[-1] = np.cos(theta_2)

        return (~adj) @ ico.element(q, 'Q') @ adj

    return _map


#############################################
# Generate irreps
#############################################

from joblib import Memory
from escnn.group import __cache_path__
cache = Memory(__cache_path__, verbose=2)

@cache.cache(ignore=['ico'])
def _build_ico_irrep(ico: Icosahedral, l: int):
    
    if l == 3:
        
        # Representation of the generator of the cyclic subgroup of order 5
        rho_p = np.zeros((3, 3))
        
        rho_p[0, 0] = rho_p[1, 1] = np.cos(144 / 180. * np.pi)
        rho_p[1, 0] = np.sin(144 / 180. * np.pi)
        rho_p[0, 1] = -np.sin(144 / 180. * np.pi)
        rho_p[2, 2] = 1.
        
        # Representation of the generator of the cyclic subgroup of order 2
        rho_q = np.zeros((3, 3))
        rho_q[0, 0] = 1. / np.sqrt(5)
        rho_q[0, 2] = - 2. / np.sqrt(5)
        rho_q[1, 1] = - 1
        rho_q[2, 0] = - 2. / np.sqrt(5)
        rho_q[2, 2] = - 1. / np.sqrt(5)
        
        generators = [
            (ico._generators[0], rho_p),
            (ico._generators[1], rho_q),
        ]
        
        return generate_irrep_matrices_from_generators(ico, generators)

    elif l == 4:

        # Representation of the generator of the cyclic subgroup of order 5
        rho_p = np.zeros((4, 4))

        rho_p[0, 0] = rho_p[1, 1] = np.cos(72 / 180. * np.pi)
        rho_p[1, 0] = np.sin(72 / 180. * np.pi)
        rho_p[0, 1] = -np.sin(72 / 180. * np.pi)

        rho_p[2, 2] = rho_p[3, 3] = np.cos(144 / 180. * np.pi)
        rho_p[3, 2] = np.sin(144 / 180. * np.pi)
        rho_p[2, 3] = -np.sin(144 / 180. * np.pi)

        # Representation of the generator of the cyclic subgroup of order 2
        rho_q = np.zeros((4, 4))
        rho_q[0, 2] = -1
        rho_q[1, 1] = 2. / np.sqrt(5)
        rho_q[1, 3] = 1. / np.sqrt(5)
        rho_q[2, 0] = -1
        rho_q[3, 1] = 1. / np.sqrt(5)
        rho_q[3, 3] = -2. / np.sqrt(5)

        generators = [
            (ico._generators[0], rho_p),
            (ico._generators[1], rho_q),
        ]

        return generate_irrep_matrices_from_generators(ico, generators)
    else:
        raise ValueError()





import numpy as np
import math


def psi(theta: float, k: int = 1, gamma: float = 0.):
    r"""
    Rotation matrix corresponding to the angle :math:`k \theta + \gamma`.
    """
    x = k * theta + gamma
    c, s = np.cos(x), np.sin(x)
    return np.array(([
        [c, -s],
        [s,  c],
    ]))


def chi(s: int):
    #
    r"""
    The orthonormal matrix associated to the reflection along the :math:`y` axis if ``s=1``, the identity otherwise.
    """

    assert s in [0, 1]
    s = (-1 if s else 1)
    # assert s in [-1, 1]
    return np.array(([
        [1, 0],
        [0, s],
    ]))


def psichi(theta: float, s: int, k: int = 1, gamma: float = 0.):
    r"""
    Rotation matrix corresponding to the angle :math:`k \theta + \gamma` if `s=0`.
    Otherwise, it corresponds to the reflection along the axis defined by that angle.
    
    It is equal to::
        
        psi(theta, k, gamma) @ chi(s)
    
    """

    assert s in [0, 1]
    s = (-1 if s else 1)
    # assert s in [-1, 1]
    
    x = k * theta + gamma
    return np.array(([
        [np.cos(x), -s*np.sin(x)],
        [np.sin(x),  s*np.cos(x)],
    ]))


def cycle_isclose(a, b, S, rtol=1e-9, atol=1e-11):
    r"""
    
    Cyclic "isclose" check.
    
    Checks if the numbers ``a`` and ``b`` are close to each other in a cycle of length ``S``,
    i.e. if ``a - b`` is close to a multiple of ``S``.
    
    """
    
    d = (a - b) % S
    
    close_0 = math.isclose(d, 0., rel_tol=rtol, abs_tol=atol) and d >= 0.
    close_S = math.isclose(d, S, rel_tol=rtol, abs_tol=atol) and d <= S
    
    return close_0 or close_S




from __future__ import annotations

import escnn.kernels
import escnn.group

from .utils import linear_transform_array_nd

from abc import ABC, abstractmethod
from typing import Tuple, Callable, List, Union

from collections import defaultdict

import numpy as np


__all__ = ["GSpace"]


class GSpace(ABC):
    
    def __init__(self, fibergroup: escnn.group.Group, dimensionality: int, name: str):
        r"""
        Abstract class for G-spaces.
        
        A ``GSpace`` describes the space where a signal lives (e.g. :math:`\R^2` for planar images) and its symmetries
        (e.g. rotations or reflections).
        As an `Euclidean` base space is assumed, a G-space is fully specified by the ``dimensionality`` of the space
        and a choice of origin-preserving symmetry group (``fibergroup``).
        
        .. seealso::
            
            :class:`~escnn.gspaces.GSpace0D`,
            :class:`~escnn.gspaces.GSpace2D`,
            :class:`~escnn.gspaces.GSpace3D`,
            or the factory methods
            :class:`~escnn.gspaces.flipRot3dOnR3`,
            :class:`~escnn.gspaces.rot3dOnR3`,
            :class:`~escnn.gspaces.rot2dOnR3`,
            :class:`~escnn.gspaces.flipRot2dOnR2`,
            :class:`~escnn.gspaces.rot2dOnR2`,
            :class:`~escnn.gspaces.flip2dOnR2`,
            :class:`~escnn.gspaces.trivialOnR2`
        
        .. note ::
        
            Mathematically, this class describes a *Principal Bundle*
            :math:`\pi : (\R^D, +) \rtimes G \to \mathbb{R}^D, tg \mapsto tG`,
            with the Euclidean space :math:`\mathbb{R}^D` (where :math:`D` is the ``dimensionality``) as `base space`
            and :math:`G` as `fiber group` (``fibergroup``).
            For more details on this interpretation we refer to
            `A General Theory of Equivariant CNNs On Homogeneous Spaces <https://papers.nips.cc/paper/9114-a-general-theory-of-equivariant-cnns-on-homogeneous-spaces.pdf>`_.
        
        
        Args:
            fibergroup (Group): the fiber group
            dimensionality (int): the dimensionality of the Euclidean space on which a signal is defined
            name (str): an identification name
        
        Attributes:
            ~.fibergroup (Group): the fiber group
            ~.dimensionality (int): the dimensionality of the Euclidean space on which a signal is defined
            ~.name (str): an identification name
            ~.basespace (str): the name of the space whose symmetries are modeled. It is an Euclidean space :math:`\R^D`.
        
        """
        
        # TODO move this sub-package to PyTorch

        self.name = name
        self.dimensionality = dimensionality
        self.fibergroup = fibergroup
        self.basespace = f"R^{self.dimensionality}"

        # in order to not recompute the basis for the same intertwiner as many times as it appears, we store the basis
        # in these dictionaries the first time we compute it

        # Store the computed intertwiners between irreps
        # - key = (filter size, sigma, rings)
        # - value = dictionary mapping (input_irrep, output_irrep) pairs to the corresponding basis
        self._irreps_intertwiners_basis_memory = defaultdict(lambda: dict())

        # Store the computed intertwiners between general representations
        # - key = (filter size, sigma, rings)
        # - value = dictionary mapping (input_repr, output_repr) pairs to the corresponding basis
        self._fields_intertwiners_basis_memory = defaultdict(dict)

        # Store the computed intertwiners between general representations
        # - key = (input_repr, output_repr)
        # - value = the corresponding basis
        self._fields_intertwiners_basis_memory_fiber_basis = dict()

    def type(self, *representations: escnn.group.Representation) -> escnn.nn.FieldType:
        f"""
            Shortcut to build a :class:`~escnn.nn.FieldType`.
            This is equivalent to ```FieldType(gspace, representations)```.
        """
        return escnn.nn.FieldType(self, representations)

    @abstractmethod
    def restrict(self, id) -> Tuple[GSpace, Callable, Callable]:
        r"""

        Build the :class:`~escnn.gspaces.GSpace` associated with the subgroup of the current fiber group identified by
        the input ``id``.
        This reduces the level of symmetries of the base space to be considered.

        Check the ``restrict`` method's documentation in the non-abstract subclass used for a description of the
        parameter ``id``.

        Args:
            id: id of the subgroup

        Returns:
            a tuple containing

                - **gspace**: the restricted gspace

                - **back_map**: a function mapping an element of the subgroup to itself in the fiber group of the original space

                - **subgroup_map**: a function mapping an element of the fiber group of the original space to itself in the subgroup (returns ``None`` if the element is not in the subgroup)

        """
        pass

    def featurefield_action(self, input: np.ndarray, repr: escnn.group.Representation, element: escnn.group.GroupElement, order: int =2) -> np.ndarray:
        r"""
        
        This method implements the action of the symmetry group on a feature field defined over the basespace of this
        G-space.
        It includes both an action over the basespace (e.g. a rotation of the points on the plane) and a transformation
        of the channels by left-multiplying them with a representation of the fiber group.

        The method takes as input a tensor (``input``), a representation (``repr``) and an ``element`` of the fiber
        group. The tensor ``input`` is the feature field to be transformed and needs to be compatible with this G-space
        and the representation (i.e. its number of channels equals the size of that representation).
        ``element`` needs to belong to the fiber group: check :meth:`escnn.group.Group.is_element`.
        This method returns a transformed tensor through the action of ``element``.

        More precisely, given an input tensor, interpreted as an :math:`c`-dimensional signal
        :math:`f: \R^D \to \mathbb{R}^c` defined over the base space :math:`\R^D`, a representation
        :math:`\rho: G \to \mathbb{R}^{c \times c}` of :math:`G` and an element :math:`g \in G` of the fiber group,
        the method returns the transformed signal :math:`f'` defined as:

        .. math::
            f'(x) := \rho(g) f(g^{-1} x)

        .. note ::

            Mathematically, this method transforms the input with the **induced representation** from the input ``repr``
            (:math:`\rho`) of the symmetry group (:math:`G`) to the *total space* (:math:`P`), i.e.
            with :math:`Ind_{G}^{P} \rho`.
            For more details on this, see
            `General E(2)-Equivariant Steerable CNNs <https://arxiv.org/abs/1911.08251>`_ or
            `A General Theory of Equivariant CNNs On Homogeneous Spaces <https://papers.nips.cc/paper/9114-a-general-theory-of-equivariant-cnns-on-homogeneous-spaces.pdf>`_.

        Args:
            input (~numpy.ndarray): input tensor
            repr (Representation): representation of the fiber group
            element (GroupElement): element of the fiber group

        Returns:
            the transformed tensor

        """
        
        from warnings import warn
        warn('Use `escnn.nn.FieldType.transform()` instead', DeprecationWarning)
        
        assert repr.group == self.fibergroup
        assert element.group == self.fibergroup

        rho = repr(element)
    
        output = np.einsum("oi,bi...->bo...", rho, input)
    
        return self._interpolate_transform_basespace(output, element, order=order)

    @property
    @abstractmethod
    def basespace_action(self) -> escnn.group.Representation:
        r"""

        Defines how the fiber group transforms the base space.

        More precisely, this method defines how an element :math:`g \in G` of the fiber group transforms a point
        :math:`x \in X \cong \R^d` of the base space.
        This action is defined as a :math:`d`-dimensional linear :class:`~escnn.group.Representation` of :math:`G`.
        
        """
        pass

    def _interpolate_transform_basespace(
            self,
            input: np.ndarray,
            element: escnn.group.GroupElement,
            order: int = 2,
    ) -> np.ndarray:
        r"""

        Defines how the fiber group transforms the base space.

        The methods takes a tensor compatible with this space (i.e. whose spatial dimensions are supported by the
        base space) and returns the transformed tensor.

        More precisely, given an input tensor, interpreted as an :math:`n`-dimensional signal
        :math:`f: X \to \mathbb{R}^n` defined over the base space :math:`X`, and an element :math:`g \in G` of the
        fiber group, the methods return the transformed signal :math:`f'` defined as:

        .. math::
            f'(x) := f(g^{-1} x)

        This method is specific of the particular GSpace and defines how :math:`g^{-1}` transforms a point
        :math:`x \in X` of the base space.


        Args:
            input (~numpy.ndarray): input tensor
            element (GroupElement): element of the fiber group

        Returns:
            the transformed tensor

        """
        assert element.group == self.fibergroup
        action = self.basespace_action
        trafo = action(element)
        return linear_transform_array_nd(input, trafo, order=order)

    @property
    def irreps(self) -> List[escnn.group.IrreducibleRepresentation]:
        r"""
        list containing all the already built irreducible representations of the fiber group of this space.

        .. seealso::

            See :attr:`escnn.group.Group.irreps` for more details

        """
        return self.fibergroup.irreps()

    @property
    def representations(self):
        r"""
        Dictionary containing all the already built representations of the fiber group of this space.

        .. seealso::

            See :attr:`escnn.group.Group.representations` for more details

        """
        return self.fibergroup.representations

    @property
    def trivial_repr(self) -> escnn.group.Representation:
        r"""
        The trivial representation of the fiber group of this space.

        .. seealso::

            :attr:`escnn.group.Group.trivial_representation`

        """
        return self.fibergroup.trivial_representation

    def irrep(self, *id) -> escnn.group.IrreducibleRepresentation:
        r"""
        Builds the irreducible representation (:class:`~escnn.group.IrreducibleRepresentation`) of the fiber group
        identified by the input arguments.

        .. seealso::

            This method is a wrapper for :meth:`escnn.group.Group.irrep`. See its documentation for more details.
            Check the documentation of :meth:`~escnn.group.Group.irrep` of the specific fiber group used for more
            information on the valid ``id``.


        Args:
            *id: parameters identifying the irrep.

        """
        return self.fibergroup.irrep(*id)

    @property
    def regular_repr(self) -> escnn.group.Representation:
        r"""
        The regular representation of the fiber group of this space.

        .. seealso::

            :attr:`escnn.group.Group.regular_representation`

        """
        return self.fibergroup.regular_representation

    def quotient_repr(self, subgroup_id) -> escnn.group.Representation:
        r"""
        Builds the quotient representation of the fiber group of this space with respect to the subgroup identified
        by ``subgroup_id``.
        
        Check the :meth:`~escnn.gspaces.GSpace.restrict` method's documentation in the non-abstract subclass used
        for a description of the parameter ``subgroup_id``.

        .. seealso::
            
            See :attr:`escnn.group.Group.quotient_representation` for more details on the representation.
        
        Args:
            subgroup_id: identifier of the subgroup

        """
        return self.fibergroup.quotient_representation(subgroup_id)

    def induced_repr(self, subgroup_id, repr: escnn.group.Representation) -> escnn.group.Representation:
        r"""
        Builds the induced representation of the fiber group of this space from the representation ``repr`` of
        the subgroup identified by ``subgroup_id``.

        Check the :meth:`~escnn.gspaces.GSpace.restrict` method's documentation in the non-abstract subclass used
        for a description of the parameter ``subgroup_id``.

        .. seealso::
            
            See :attr:`escnn.group.Group.induced_representation` for more details on the representation.
        
        Args:
            subgroup_id: identifier of the subgroup
            repr (Representation): the representation of the subgroup to induce

        """
        return self.fibergroup.induced_representation(subgroup_id, repr)

    @property
    def testing_elements(self):
        return self.fibergroup.testing_elements()

    def build_fiber_intertwiner_basis(self,
                                      in_repr: escnn.group.Representation,
                                      out_repr: escnn.group.Representation,
                                      ) -> escnn.kernels.KernelBasis:
        r"""


        Args:
            in_repr (Representation): the input representation
            out_repr (Representation): the output representation

        Returns:
            the analytical basis

        """
    
        assert isinstance(in_repr, escnn.group.Representation)
        assert isinstance(out_repr, escnn.group.Representation)
    
        assert in_repr.group == self.fibergroup
        assert out_repr.group == self.fibergroup
    
        if (in_repr.name, out_repr.name) not in self._fields_intertwiners_basis_memory_fiber_basis:
        
            basis = escnn.kernels.kernels_on_point(in_repr, out_repr)
        
            # store the basis in the dictionary
            self._fields_intertwiners_basis_memory_fiber_basis[(in_repr.name, out_repr.name)] = basis
    
        # return the dictionary with all the basis built
        return self._fields_intertwiners_basis_memory_fiber_basis[(in_repr.name, out_repr.name)]

    def build_kernel_basis(self,
                           in_repr: escnn.group.Representation,
                           out_repr: escnn.group.Representation,
                           sigma: Union[float, List[float]],
                           rings: List[float],
                           **kwargs) -> escnn.kernels.KernelBasis:
        r"""

        Builds a basis for the space of the equivariant kernels with respect to the symmetries described by this
        :class:`~escnn.gspaces.GSpace`.

        A kernel :math:`\kappa` equivariant to a group :math:`G` needs to satisfy the following equivariance constraint:

        .. math::
            \kappa(gx) = \rho_\text{out}(g) \kappa(x) \rho_\text{in}(g)^{-1}  \qquad \forall g \in G, x \in \R^D

        where :math:`\rho_\text{in}` is ``in_repr`` while :math:`\rho_\text{out}` is ``out_repr``.


        Because the equivariance constraints only restrict the angular part of the kernels, any radial profile is
        permitted.
        The basis for the radial profile used here contains rings with different radii (``rings``)
        associated with (possibly different) widths (``sigma``).
        A ring is implemented as a Gaussian function over the radial component, centered at one radius
        (see also :class:`~escnn.kernels.GaussianRadialProfile`).

        .. note ::
            This method is a wrapper for the functions building the bases which are defined in :doc:`escnn.kernels`:

            - :meth:`escnn.kernels.kernels_O2_act_R2`,

            - :meth:`escnn.kernels.kernels_SO2_act_R2`,

            - :meth:`escnn.kernels.kernels_DN_act_R2`,

            - :meth:`escnn.kernels.kernels_CN_act_R2`,

            - :meth:`escnn.kernels.kernels_Flip_act_R2`,

            - :meth:`escnn.kernels.kernels_Trivial_act_R2`


        Args:
            in_repr (Representation): the input representation
            out_repr (Representation): the output representation
            sigma (list or float): parameters controlling the width of each ring of the radial profile.
                    If only one scalar is passed, it is used for all rings
            rings (list): radii of the rings defining the radial profile
            **kwargs: Group-specific keywords arguments for ``_basis_generator`` method

        Returns:
            an instance of :class:`~escnn.kernels.KernelBasis` representing the analytical basis

        """
        
        # TODO - move sigma and rings in kwargs
        
        # TODO - solve incompatibility with subclasses' method signature
        # now different subgroups have different keys, so no point is sharing this piece of code
        # maybe use an 'parse_args(**kwargs)'  abstract method????
        # same for _basis_generator
    
        assert isinstance(in_repr, escnn.group.Representation)
        assert isinstance(out_repr, escnn.group.Representation)
    
        assert in_repr.group == self.fibergroup
        assert out_repr.group == self.fibergroup
    
        if isinstance(sigma, float):
            sigma = [sigma] * len(rings)
    
        assert all([s > 0. for s in sigma])
        assert len(sigma) == len(rings)
    
        # build the key
        key = dict(**kwargs)
        key["sigma"] = tuple(sigma)
        key["rings"] = tuple(rings)
        key = tuple(sorted(key.items()))
    
        if (in_repr.name, out_repr.name) not in self._fields_intertwiners_basis_memory[key]:
            # TODO - we could use a flag in the args to choose whether to store it or not
        
            basis = self._basis_generator(in_repr, out_repr, rings, sigma, **kwargs)
        
            # store the basis in the dictionary
            self._fields_intertwiners_basis_memory[key][(in_repr.name, out_repr.name)] = basis
    
        # return the dictionary with all the basis built for this filter size
        return self._fields_intertwiners_basis_memory[key][(in_repr.name, out_repr.name)]

    @abstractmethod
    def _basis_generator(self,
                         in_repr: escnn.group.Representation,
                         out_repr: escnn.group.Representation,
                         rings: List[float],
                         sigma: List[float],
                         **kwargs):
        pass

    def __repr__(self):
        return self.name




from escnn.gspaces import *
from escnn.group import *
from escnn.nn import GeometricTensor
from .equivariant_module import EquivariantModule

from escnn.kernels import HarmonicPolynomialR3Generator

import torch
import numpy as np

from typing import Tuple, Union


__all__ = ["HarmonicPolynomialR3"]


class HarmonicPolynomialR3(EquivariantModule):

    def __init__(self, L: int, group: Union[Group, str] = 'so3'):
        r"""
            Module which computes the *harmonic polynomials* in :math:`\R^3` up to order `L`.

            The argument `group` can be a string (`"so3"` or `"o3"`) or a group (instance of
            :class:`~escnn.group.SO3` or :class:`~escnn.group.O3` ).

            This equivariant module takes a set of 3-dimensional points transforming according to
            the :meth:`~escnn.group.SO3.standard_representation` of :math:`SO(3)`
            (or the :meth:`~escnn.group.O3.standard_representation` of :math:`O(3)`)
            and outputs :math:`(L+1)^2` dimensional
            feature vectors transforming like spherical harmonics according to
            :meth:`~escnn.group.SO3.bl_sphere_representation` of :math:`SO(3)`
            (or :meth:`~escnn.group.O3.bl_sphere_representation` of :math:`O(3)`)
            with `L=L`.

            .. seealso ::

                Harmonic polynomial are related to the spherical harmonics.
                Check the
                `Wikipedia page <https://en.wikipedia.org/wiki/Spherical_harmonics#Harmonic_polynomial_representation>`_
                about them.

        """

        super(HarmonicPolynomialR3, self).__init__()

        if group == 'so3':
            group = so3_group(L)
        elif group == 'o3':
            group = o3_group(L)

        assert isinstance(group, SO3) or isinstance(group, O3), group

        self.G: Union[SO3, O3] = group

        self.gspace = no_base_space(self.G)

        self.L = L

        self.harmonics_generator = HarmonicPolynomialR3Generator(self.L)
        self.rho = self.G.bl_sphere_representation(L)

        self.in_type = self.gspace.type(self.G.standard_representation())
        self.out_type = self.gspace.type(self.rho)

    def forward(self, points: GeometricTensor):
        assert points.type == self.in_type
        features = self.harmonics_generator(points.tensor)
        return self.out_type(features, coords=points.coords)

    def evaluate_output_shape(self, input_shape: Tuple) -> Tuple:
        assert len(input_shape) == 2
        assert input_shape[1] == self.in_type.size

        return (input_shape[0], self.out_type.size)

    def check_equivariance(self, atol: float = 1e-4, rtol: float = 1e-3):

        device = self.harmonics_generator.cob_1.device

        N = 40
        points = torch.randn(N, 3, device=device)
        # radii = torch.norm(points, dim=-1).view(-1, 1)
        # points = points / radii
        # points[radii.view(-1) < 1e-3, :] = 0.

        points = self.in_type(points)

        sh = self(points)
        for _ in range(10):
            g = self.G.sample()
            sh_rot = self(g @ points)
            rot_sh = g @ sh
            assert torch.allclose(rot_sh.tensor, sh_rot.tensor, atol=atol, rtol=rtol), (rot_sh - sh_rot).tensor.abs().max().item()





import torch

from escnn.nn import GeometricTensor
from escnn.nn import FieldType
from .equivariant_module import EquivariantModule


from typing import List, Tuple, Any, Union, Dict

__all__ = ["MergeModule"]


class MergeModule(EquivariantModule):
    
    def __init__(self,
                 modules: List[Tuple[EquivariantModule, Union[str, List[str]]]],
                 ):
        r"""
        
        Applies different modules to multiple tensors in input.
        
        ``modules`` contains a list of pairs, each containing an :class:`~escnn.nn.EquivariantModule` and a label (or a
        list of labels).
        
        This module takes as input a dictionary mapping labels to tensors.
        Then, each module in ``modules`` is applied to the tensors associated to its labels.
        Finally, output tensors are stacked together.
        
        .. todo ::
            Technically this is not an EquivariantModule as the input is not a single tensor.
            Either fix EquivariantModule to support multiple inputs and outputs or set this as just subclass of
            torch.nn.Module.
        
        Args:
            modules (list): list of modules to apply to the labeled input tensors
            
        """

        super(MergeModule, self).__init__()
        
        labels = []
        
        for i in range(len(modules)):
            if isinstance(modules[i][1], str):
                modules[i] = (modules[i][0], [modules[i][1]])
            else:
                assert isinstance(modules[i][1], list)
                for s in modules[i][1]:
                    assert isinstance(s, str)

            labels += modules[i][1]
        
        self.in_type = None
        self.gspace = modules[0][0].in_type.gspace
        
        out_repr = []
        
        for module, labels in modules:
            if isinstance(module.in_type, tuple):
                assert all(t.gspace == self.gspace for t in module.in_type)
            else:
                assert module.in_type.gspace == self.gspace
                
            out_repr += module.out_type.representations
        
        self.out_type = FieldType(self.gspace, out_repr)
        
        self._labels = []
        # add the input modules as sub-modules
        for i, (module, labels) in enumerate(modules):
            self._labels.append(labels)
            self.add_module('submodule_{}'.format(i), module)
        
    def forward(self, input: Dict[str, GeometricTensor]) -> GeometricTensor:
        r"""
        
        Apply each module to the corresponding input tensors and stack the results
        
        Args:
            input (dict): a dictionary mapping each label to a GeometricTensor

        Returns:
            the concatenation of the output of each module
            
        """
        
        # compute the output shape
        out_shape = self.evaluate_output_shape(**{l: t.tensor.shape for l, t in input.items()})
        
        device = list(input.values())[0].tensor.device
        
        # pre-allocate the output tensor
        output = torch.empty(out_shape, dtype=torch.float, device=device)
        
        last_channel = 0
        out_0 = None
        
        # iterate through the modules
        for i, labels in enumerate(self._labels):
            module = getattr(self, f"submodule_{i}")
            # retrieve the corresponding sub-tensor
            out_i = module(*[input[l] for l in labels])
            if out_0 is None:
                out_0 = out_i
            
            assert out_0.has_same_coords(out_i)
            
            output[:, last_channel:last_channel + module.out_type.size, ...] = out_i.tensor
            last_channel += module.out_type.size
        
        return GeometricTensor(output, self.out_type, out_0.coords)

    def evaluate_output_shape(self, **input_shapes: Tuple[int, ...]) -> Tuple[int, ...]:
    
        output_shapes = []
    
        # iterate through the modules
        for i, labels in enumerate(self._labels):
            module = getattr(self, f"submodule_{i}")
            # evaluate the corresponding output shape
            # output_shapes.append(module.evaluate_output_shape(*[input_shapes[l] for l in labels]))
            os = module.evaluate_output_shape(*[input_shapes[l] for l in labels])
            output_shapes.append(list(os))
    
        out_shape = list(output_shapes[0])
    
        for os in output_shapes[1:]:
            assert out_shape[0] == os[0]
            assert out_shape[2:] == os[2:]
            
            out_shape[1] += os[1]
        
        return out_shape

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
        pass


from lie_learn.representations.SO3.wigner_d import wigner_D_matrix

from scipy.spatial.transform import Rotation

import numpy as np

from ._numerical_thomson import thomson_sphere, thomson_so3, thomson_cube_sphere, thomson_cube_so3

import warnings


__all__ = [
    'PARAMETRIZATION',
    'PARAMETRIZATIONS',
    'IDENTITY',
    '_change_param',
    '_check_param',
    '_invert',
    '_combine',
    '_equal',
    '_hash',
    '_repr',
    '_wigner_d_matrix',
    '_character',
    '_grid',
    '_sphere_grid',
    'find_rotation_pole2point',
]

PARAMETRIZATION = "Q"

PARAMETRIZATIONS = [
    'Q',
    'MAT',
    'EV',
    
    # estrinsic rotations
    'xyz',
    'xzy',
    'yxz',
    'yzx',
    'zxy',
    'zyx',
    
    'xyx',
    'xzx',
    'yxy',
    'yzy',
    'zxz',
    'zyz',
    
    # intrinsic rotations
    'XYZ',
    'XZY',
    'YXZ',
    'YZX',
    'ZXY',
    'ZYX',
    
    'XYX',
    'XZX',
    'YXY',
    'YZY',
    'ZXZ',
    'ZYZ',
]


def _to_scipy_rot(element, param: str) -> Rotation:
    assert param in PARAMETRIZATIONS
    if param == 'Q':
        assert isinstance(element, np.ndarray)
        assert element.shape == (4,)
        assert np.isclose(np.linalg.norm(element), 1.)

        return Rotation.from_quat(element)
    elif param == 'MAT':
        assert isinstance(element, np.ndarray)
        assert element.shape == (3, 3)
        assert np.isclose(np.linalg.det(element), 1.)

        return Rotation.from_matrix(element)
    elif param == 'EV':
        return Rotation.from_rotvec(element)
    else:
        return Rotation.from_euler(param, element)


def _from_scipy_rot(element: Rotation, param: str) -> np.ndarray:
    assert param in PARAMETRIZATIONS
    if param == 'Q':
        return element.as_quat()
    elif param == 'MAT':
        return element.as_matrix()
    elif param == 'EV':
        return element.as_rotvec()
    else:
        with warnings.catch_warnings():
            # to prevent printing warnings about Gimbal-Lock
            warnings.simplefilter("ignore")

            return element.as_euler(param)


def _change_param(element, p_from: str, p_to: str):
    
    if p_from == p_to:
        return element
    
    rot = _to_scipy_rot(element, p_from)
    
    return _from_scipy_rot(rot, p_to)


def _check_param(element, param):
    try:
        _to_scipy_rot(element, param)
    except:
        return False
    
    return True


def _invert(element, param=PARAMETRIZATION):
    
    return _from_scipy_rot(
        _to_scipy_rot(element, param).inv(),
        param
    )


def _combine(e1, e2,
             param=PARAMETRIZATION,
             param1=None,
             param2=None
             ):
    r"""
        Compute `e1 * e2`.
        This corresponds to a first rotation by `e2` followed by a rotation by `e1`
    
    """
    
    if param1 is None:
        param1 = param
    if param2 is None:
        param2 = param

    return _from_scipy_rot(
        _to_scipy_rot(e1, param1) * _to_scipy_rot(e2, param2),
        param
    )

def _equal(e1, e2,
           param: str = PARAMETRIZATION,
           param1: str = None,
           param2: str = None,
           atol: float = 1e-7,
           rtol: float = 1e-5
           ) -> bool:
    
    if param1 is None:
        param1 = param
    if param2 is None:
        param2 = param
    
    if not isinstance(e1, np.ndarray):
        e1 = np.asarray(e1).reshape(1, -1)
    if not isinstance(e2, np.ndarray):
        e2 = np.asarray(e2).reshape(1, -1)
    
    # convert the rotations to quaternion
    q1 = _to_scipy_rot(e1, param1).as_quat()
    q2 = _to_scipy_rot(e2, param2).as_quat()
    
    # quaternions are a double cover of SO(3)
    # as both q and -q correspond to the same element of SO(3)
    # therefore, we need to check both cases
    return np.allclose(q1, q2, rtol=rtol, atol=atol) or np.allclose(-q1, q2, rtol=rtol, atol=atol)


def _hash(element, param: str = PARAMETRIZATION):
    element = np.around(_change_param(element, param, 'MAT').reshape(-1), 4)
    return hash(tuple(element))


def _repr(element, param: str = PARAMETRIZATION) -> str:
    element = _change_param(element, param, 'MAT')
    return '\n'.join(element.__repr__()[7:-2].split('\n       '))
    # element = _change_param(element, param, 'Q')
    # return '\n'.join(element.__repr__()[6:-1].split('\n       '))


def _wigner_d_matrix(element, l, param=PARAMETRIZATION, field: str = 'real'):
    wigner_d_param = 'ZYZ'
    
    if param != wigner_d_param:
        element = _change_param(element, p_from=param, p_to=wigner_d_param)
    
    return wigner_D_matrix(l, element[0, ...], element[1, ...], element[2, ...], field=field)


def _character(element, l, param=PARAMETRIZATION):
    if l == 0:
        return 1.
    else:
        element = _change_param(element, p_from=param, p_to='Q')
        theta = 2 * np.arctan2(
            np.linalg.norm(element[..., :-1]),
            element[..., -1]
        )
        sin = np.sin(.5 * theta)
        if not np.isclose(sin, 0.):
            c = np.sin((l + .5) * theta) / np.sin(0.5 * theta)
        else:
            # c[np.sin(.5*theta) == 0.] = (2*l + 1)
            c = (2 * l + 1)
        
        return c


IDENTITY = _change_param(np.array([0., 0., 0., 1.]), p_from='Q', p_to=PARAMETRIZATION)


def find_rotation_pole2point(point: np.ndarray):

    assert point.shape == (3,)

    point = point / np.linalg.norm(point)

    rot = np.zeros((3, 3))

    rot[:, 2] = point

    y = np.array([0., 0., 1.])

    if np.isclose(np.abs(np.dot(y, point)), 1.):
        y = np.array([1., 0., 0.])

    rot[:, 1] = y - np.dot(y, point) * point
    rot[:, 1] /= np.linalg.norm(rot[:, 1])
    rot[:, 0] = np.cross(rot[:, 2], rot[:, 1])
    det = np.linalg.det(rot)
    rot[:, 0] *= det

    assert np.isclose(np.linalg.det(rot), 1.), rot

    return rot


#############################################
# GRIDS
#############################################

# Sphere grids

def _random_sphere_samples(n: int, rnd=None):
    assert n > 0
    rng = np.random.RandomState(rnd) if rnd is not None else np.random
    points = rng.randn(n, 3)
    points /= np.linalg.norm(points, axis=1, keepdims=True)
    return points


def _spherical_healpix_grid(N_side: int) -> np.ndarray:
    north_points = []

    # north polar cap
    for i in range(1, N_side):
        for j in range(1, 4 * i + 1):
            cos_theta = 1. - i ** 2 / (3 * N_side ** 2)
            phi = np.pi / (2 * i) * (j - 0.5)

            sin_theta = np.sqrt(1. - cos_theta ** 2)

            north_points.append(
                np.array([np.cos(phi) * sin_theta, np.sin(phi) * sin_theta, cos_theta])
            )

    # north equatorial belt
    for i in range(N_side, 2 * N_side + 1):
        for j in range(1, 4 * N_side + 1):
            cos_theta = 4. / 3. - 2 * i / (3 * N_side)
            s = (i - N_side + 1) % 2
            phi = np.pi / (2 * N_side) * (j - s / 2.)

            sin_theta = np.sqrt(1. - cos_theta ** 2)

            north_points.append(
                np.array([np.cos(phi) * sin_theta, np.sin(phi) * sin_theta, cos_theta])
            )

    points = []
    # add points on the south pole (symmetric to north pole wrt equator, i.e. z=0)
    for p in north_points:
        if p[2] > 0.:
            sp = p.copy()
            sp[2] *= -1
            points.append(sp)
    points += north_points

    N_pix = 12 * N_side ** 2
    assert len(points) == N_pix, (len(points), N_side, N_pix)

    points = np.stack(points, axis=0)

    return points


def _plato_sphere_samples(solid: str):
    solid = solid.split('_')
    if len(solid) == 1:
        solid = solid[0]
        points = None
    elif len(solid) == 2:
        points = solid[1]
        solid = solid[0]
    else:
        raise ValueError

    assert solid in ['ico', 'cube', 'tetra']
    assert points in [None, 'vertices', 'faces', 'edges']

    so3_grid = _grid(solid, parametrization='MAT')

    if points is None:
        origin = np.array([0., 0., 1.]).reshape(-1, 1)
    elif solid == 'cube' and points == 'vertices':
        origin = np.array([1., 1., 1.]).reshape(-1, 1)
    else:
        raise NotImplementedError

    # origin = np.array([1., 1., 1.]).reshape(-1, 1)
    origin /= np.linalg.norm(origin)

    class HashablePoint:

        def __init__(self, point: np.ndarray):
            self.p = point

        def __eq__(self, other):
            return np.allclose(self.p, other.p)

        def __hash__(self):
            element = np.around(self.p.reshape(-1), 4)
            return hash(tuple(element))

    points = set(
        HashablePoint(r @ origin) for r in so3_grid
    )
    points = np.stack([p.p.flatten() for p in points], axis=0)
    return points


def _fibonacci_sphere_samples(N: int):

    idx = np.arange(N, dtype=np.float)

    phi = np.pi * (3 - np.sqrt(5))

    y = 1. - 2*(idx / (N-1))

    radius = np.sqrt(1 - y**2)

    theta = phi * idx

    x = np.cos(theta) * radius
    z = np.sin(theta) * radius

    points = np.stack([x, y, z], axis=-1)

    return points


def _longlat_sphere_grid(N: int, M: int):

    points = []

    assert M > 1

    for la in range(M):

        if la == 0:
            points.append(np.array([0., 0., 1.]))
            continue

        if la == M-1:
            points.append(np.array([0., 0., -1.]))
            continue

        theta = la * np.pi / (M - 1)
        z = np.cos(theta)

        r = np.sqrt(1 - z ** 2)

        for lo in range(N):
            phi = lo * 2*np.pi / N

            x = np.cos(phi)*r
            y = np.sin(phi)*r

            point = np.array([x, y, z])
            points.append(point)

    points = np.stack(points, axis=0)

    return points


def _sphere_grid(type: str, *args, adj: np.ndarray = None, **kwargs) -> np.ndarray:
    if type == 'rand':
        elements = _random_sphere_samples(*args, **kwargs)
    elif type in ['ico', 'cube', 'tetra', 'cube_vertices']:
        elements = _plato_sphere_samples(type)
    elif type == 'healpix':
        elements = _spherical_healpix_grid(*args, **kwargs)
    elif type == 'fibonacci':
        elements = _fibonacci_sphere_samples(*args, **kwargs)
    elif type == 'longlat':
        elements = _longlat_sphere_grid(*args, **kwargs)
    elif type == 'thomson':
        elements = thomson_sphere(*args, **kwargs)
    elif type == 'thomson_cube':
        elements = thomson_cube_sphere(*args, **kwargs)
    else:
        raise ValueError(f'Grid type "{type}" not recognized')

    assert elements.shape[1] == 3

    if adj is not None:
        assert adj.shape == (3, 3)
        assert np.isclose(np.abs(np.linalg.det(adj)), 1.)
        elements = elements @ adj.T

    return elements

# SO(3) grids

def _so3_hopf_uniform_grid(N) -> Rotation:
    m1 = int(np.round(np.cbrt(N * np.pi)))
    N_side = int(np.round(
        np.sqrt(N / (8 * np.pi ** 2))
    ))
    
    assert m1 > 0
    assert N_side > 0
    
    sphere_points = _spherical_healpix_grid(N_side)
    so3_points = []
    
    for sp in sphere_points:
        for i in range(m1):
            # add +1/4 such that the same rotation is not sampled on the opposite point on the sphere
            
            theta = (i + .25) * 2 * np.pi / m1
            # theta = i * 2 * np.pi / m1
            
            q = np.array([sp[0], sp[1], sp[2], 0.])
            q *= np.sin(theta / 2.)
            q[3] = np.cos(theta / 2.)
            
            so3_points.append(
                Rotation.from_quat(q)
            )
    
    # Filter duplicates
    
    class HashableRot:
        
        def __init__(self, rot: Rotation):
            self.r = rot
        
        def __eq__(self, other):
            return _equal(self.r.as_quat(), other.r.as_quat(), 'Q')
        
        def __hash__(self):
            return _hash(self.r.as_quat(), 'Q')

    so3_points = set(HashableRot(r) for r in so3_points)
    so3_points = Rotation.from_quat(np.stack([r.r.as_quat().reshape(4) for r in so3_points], axis=0))
    # print(len(sphere_points), len(so3_points),  m1, N)
    
    return so3_points


def _so3_hopf_fibonacci_grid(N) -> Rotation:
    N_circle = int(np.round(np.cbrt(N * np.pi)))
    N_sphere = int(np.round(
        np.cbrt(N**2 / np.pi)
    ))

    assert N_circle > 0
    assert N_sphere > 0

    sphere_points = _fibonacci_sphere_samples(N_sphere)
    so3_points = []

    for sp in sphere_points:
        for i in range(N_circle):
            # add +1/4 such that the same rotation is not sampled on the opposite point on the sphere

            # theta = (i + .25) * 2 * np.pi / N_circle
            theta = i * 2 * np.pi / N_circle

            q = np.array([sp[0], sp[1], sp[2], 0.])
            q *= np.sin(theta / 2.)
            q[3] = np.cos(theta / 2.)

            so3_points.append(
                Rotation.from_quat(q)
            )

    # Filter duplicates

    class HashableRot:

        def __init__(self, rot: Rotation):
            self.r = rot

        def __eq__(self, other):
            return _equal(self.r.as_quat(), other.r.as_quat(), 'Q')

        def __hash__(self):
            return _hash(self.r.as_quat(), 'Q')

    so3_points = set(HashableRot(r) for r in so3_points)
    so3_points = Rotation.from_quat(np.stack([r.r.as_quat().reshape(4) for r in so3_points], axis=0))
    # print(len(sphere_points), len(so3_points),  m1, N)

    return so3_points


def _random_samples(N: int, seed=None):
    assert N > 0
    return Rotation.random(N, seed)


def _grid(type: str, *args, adj=IDENTITY, parametrization: str = PARAMETRIZATION, **kwargs) -> np.ndarray:
    if type == 'rand':
        elements = _random_samples(*args, **kwargs)
    elif type == 'tetra':
        elements = Rotation.create_group('T')
    elif type == 'cube':
        elements = Rotation.create_group('O')
    elif type == 'ico':
        elements = Rotation.create_group('I')
    elif type == 'hopf':
        elements = _so3_hopf_uniform_grid(*args, **kwargs)
    elif type == 'fibonacci':
        elements = _so3_hopf_fibonacci_grid(*args, **kwargs)
    elif type == 'thomson':
        elements = Rotation.from_quat(thomson_so3(*args, **kwargs))
    elif type == 'thomson_cube':
        elements = Rotation.from_quat(thomson_cube_so3(*args, **kwargs))
    else:
        raise ValueError(f'Grid type "{type}" not recognized')
    
    adj = _to_scipy_rot(adj, PARAMETRIZATION)
    elements = adj * elements * adj.inv()
    
    return _from_scipy_rot(elements, parametrization)


if __name__ == '__main__':
    # # plot helpix grid
    # grid = _spherical_healpix_grid(4)
    # grid = np.asarray(grid)
    
    N = 500
    # elements = _so3_hopf_uniform_grid(N)
    # elements = set(elements)
    # print(N, len(elements))
    
    # plot hoft grid
    elements = _grid('hopf', N=N, parametrization='Q')
    print(len(elements))
    elements = np.asarray(elements)
    
    grid = elements[:, :3]
    theta_2 = np.arccos(elements[:, 3]) * 2 - np.pi
    theta_2 = theta_2.reshape(-1, 1)
    grid *= theta_2
    
    from mpl_toolkits.mplot3d import Axes3D
    import matplotlib.pyplot as plt
    import numpy as np
    
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    
    ax.scatter(grid[:, 0], grid[:, 1], grid[:, 2], )
    
    plt.show()
